#include <string.h>
#include <stdio.h>
#include "Index3D_CM.h"
#ifndef MAX
#define MAX(a,b)((a)>(b)?(a):(b))
#define MIN(a,b)((a)<(b)?(a):(b))
#endif
//------------------------------------------------------------------------
// memory related functions
//------------------------------------------------------------------------
Index3D_CM::Index3D_CM(){ //default constructor
	isize=0;jsize=0;ksize=0;nsize=0; n_to_ijk=0;ijk_to_n=0;sum_nbeg=0;sum_nsize=0;
	index_nim1=0; index_nip1=0;
	index_njm1=0; index_njp1=0;
	index_nkm1=0; index_nkp1=0;
}
Index3D_CM::Index3D_CM(int isize_,int jsize_,int ksize_){//isize,jsize,ksize
	isize=0;jsize=0;ksize=0;nsize=0; n_to_ijk=0;ijk_to_n=0;sum_nbeg=0;sum_nsize=0;
	index_nim1=0; index_nip1=0;
	index_njm1=0; index_njp1=0;
	index_nkm1=0; index_nkp1=0;
	resize(isize_,jsize_,ksize_);
}
Index3D_CM::~Index3D_CM(){ //destructor
	resize(0,0,0);
}
Index3D_CM::Index3D_CM(const Index3D_CM& V){//copy constructor
	isize=0;jsize=0;ksize=0;nsize=0; n_to_ijk=0;ijk_to_n=0;sum_nbeg=0;sum_nsize=0;
	index_nim1=0; index_nip1=0;
	index_njm1=0; index_njp1=0;
	index_nkm1=0; index_nkp1=0;
	resize(V.isize  ,V.jsize    ,V.ksize);
	memcpy(index_nim1,V.index_nim1,sizeof(int)*nsize);
	memcpy(index_nip1,V.index_nip1,sizeof(int)*nsize);
	memcpy(index_njm1,V.index_njm1,sizeof(int)*nsize);
	memcpy(index_njp1,V.index_njp1,sizeof(int)*nsize);
	memcpy(index_nkm1,V.index_nkm1,sizeof(int)*nsize);
	memcpy(index_nkp1,V.index_nkp1,sizeof(int)*nsize);
	memcpy(n_to_ijk  ,V.n_to_ijk  ,sizeof(int)*nsize);
	memcpy(sum_nbeg ,V.sum_nbeg ,sizeof(int)*(isize+jsize+ksize-3));
	memcpy(sum_nsize,V.sum_nsize,sizeof(int)*(isize+jsize+ksize-3));
	for(int i=0;i<isize;i++)
	for(int j=0;j<jsize;j++)
		memcpy(ijk_to_n[i][j],V.ijk_to_n[i][j],sizeof(int)*ksize);
}

void Index3D_CM::operator=(const Index3D_CM& V){//copy operator
	resize(V.isize,V.jsize,V.ksize);
	resize(V.isize  ,V.jsize    ,V.ksize);
	memcpy(index_nim1,V.index_nim1,sizeof(int)*nsize);
	memcpy(index_nip1,V.index_nip1,sizeof(int)*nsize);
	memcpy(index_njm1,V.index_njm1,sizeof(int)*nsize);
	memcpy(index_njp1,V.index_njp1,sizeof(int)*nsize);
	memcpy(index_nkm1,V.index_nkm1,sizeof(int)*nsize);
	memcpy(index_nkp1,V.index_nkp1,sizeof(int)*nsize);
	memcpy(n_to_ijk  ,V.n_to_ijk  ,sizeof(int)*nsize);
	memcpy(sum_nbeg ,V.sum_nbeg ,sizeof(int)*(isize+jsize+ksize-3));
	memcpy(sum_nsize,V.sum_nsize,sizeof(int)*(isize+jsize+ksize-3));
	for(int i=0;i<isize;i++)
	for(int j=0;j<jsize;j++)
		memcpy(ijk_to_n[i][j],V.ijk_to_n[i][j],sizeof(int)*ksize);
}

void Index3D_CM::resize(int isize_,int jsize_,int ksize_){
	//nothing to do
	if(isize==isize_ &&
	   jsize==jsize_ &&
	   ksize==ksize_ ) return;
	else {
		// dealloc the current
		if(nsize>0) {
			delete[] n_to_ijk ; n_to_ijk =0;
			delete[] sum_nbeg ; sum_nbeg =0;
			delete[] sum_nsize; sum_nsize=0;
			delete[] index_nim1; delete[] index_nip1;
			delete[] index_njm1; delete[] index_njp1;
			delete[] index_nkm1; delete[] index_nkp1;
			for(int i=0;i<isize;i++){
			for(int j=0;j<jsize;j++) delete[] ijk_to_n[i][j];
				                     delete[] ijk_to_n[i]   ;}
									 delete[] ijk_to_n      ; ijk_to_n=0;
		}
		// alloc the newly requested
		isize=isize_;
		jsize=jsize_;
		ksize=ksize_; nsize=isize*jsize*ksize;
		if(nsize>0) {
			n_to_ijk  =new int [nsize*3];
			index_nim1=new int [nsize];
			index_nip1=new int [nsize];
			index_njm1=new int [nsize];
			index_njp1=new int [nsize];
			index_nkm1=new int [nsize];
			index_nkp1=new int [nsize];
			sum_nbeg  =new int [isize+jsize+ksize-2];
			sum_nsize =new int [isize+jsize+ksize-2];
			                          ijk_to_n       = new int**[isize];
			for(int i=0;i<isize;i++){ ijk_to_n[i]    = new int* [jsize];
			for(int j=0;j<jsize;j++)  ijk_to_n[i][j] = new int  [ksize];}
			// match the two indices
			int im=0, iM=isize-1;
			int jm=0, jM=jsize-1;
			int km=0, kM=ksize-1; int n=0;
			for(int sum=im+jm+km;sum<=iM+jM+kM;sum++) {
				int imin = MAX(im,sum-jM-kM); sum_nbeg[sum]=n; int _nsize=0;
				int imax = MIN(iM,sum-im-km);
				for(int i=imin;i<=imax;i++) {
					int jmin = MAX(jm,sum-i-kM);
					int jmax = MIN(jM,sum-i-km);
					for(int j=jmin;j<=jmax;j++) {
						int k = sum-i-j;
						n_to_ijk[3*n  ]=i; 
						n_to_ijk[3*n+1]=j; 
						n_to_ijk[3*n+2]=k;
						ijk_to_n[i][j][k]=n; n++; _nsize++;
					}
				}
				sum_nsize[sum]=_nsize;
			}
			// neighborhood
			for(int n=0;n<nsize;n++){
				int i = n_to_ijk[3*n  ];
				int j = n_to_ijk[3*n+1];
				int k = n_to_ijk[3*n+2];
				index_nim1[n] = (i==      0)?-1:ijk_to_n[i-1][j][k];
				index_nip1[n] = (i==isize-1)?-1:ijk_to_n[i+1][j][k];
				index_njm1[n] = (j==      0)?-1:ijk_to_n[i][j-1][k];
				index_njp1[n] = (j==jsize-1)?-1:ijk_to_n[i][j+1][k];
				index_nkm1[n] = (k==      0)?-1:ijk_to_n[i][j][k-1];
				index_nkp1[n] = (k==ksize-1)?-1:ijk_to_n[i][j][k+1];
			}
		}
	}
}
//------------------------------------------------------------------------
// print
//------------------------------------------------------------------------
void Index3D_CM::print_j(const double* data, const char* name, int j, int step) const {
	printf("--------------------%s :(j=%d/%d)--------------------\n", name,j,jsize);
	for(int k=ksize-1;k>=0;k-=step) {
	for(int i=0;i<isize;i+=step) printf("% .3f", data[ijk_to_n[i][j][k]]); printf("\n"); }
                                                                    printf("\n"); 
}

void Index3D_CM::print_k(const double* data, const char* name, int k, int step) const {
	printf("--------------------%s :(k=%d/%d)--------------------\n", name,k,ksize);
	for(int j=jsize-1;j>=0;j-=step) {
	for(int i=0;i<isize;i+=step) printf("% .3f", data[ijk_to_n[i][j][k]]); printf("\n"); }
	                                                                printf("\n"); 
}

