#include <stdlib.h>
#include <assert.h>
#include <omp.h>
#include "Poisson_Symmetric_7Point_CM.h"

int Poisson_Symmetric_7Point_CM::NUM_THRS;
//
void Poisson_Symmetric_7Point_CM::set_size(int isize, int jsize, int ksize){
	index3d_cm.resize(isize,jsize,ksize);
    Aiph .resize(index3d_cm.nsize); Aiph =0;
    Ajph .resize(index3d_cm.nsize); Ajph =0;	
    Akph .resize(index3d_cm.nsize); Akph =0;
    Adiag.resize(index3d_cm.nsize); Adiag=0;
    Mdiag.resize(index3d_cm.nsize); Mdiag=0;
}   
//
double Poisson_Symmetric_7Point_CM::inner_product( const VECTOR& X,
                                                   const VECTOR& Y ) {
    int nsize = index3d_cm.nsize; double sum=0;
	const double* X_=X.data;
	const double* Y_=Y.data;
#pragma omp parallel for reduction(+:sum) num_threads(NUM_THRS)
    for(int n=0;n<nsize;n++)
		sum += X_[n]*Y_[n];
	return sum;
}
void Poisson_Symmetric_7Point_CM::projection_sumzero(  VECTOR& X  )
{
	int nsize = index3d_cm.nsize; double sum=0; 
	double* X_=X.data;
#pragma omp parallel for reduction(+:sum) num_threads(NUM_THRS)
    for(int n=0;n<nsize;n++) 
		sum+=X_[n];

	double avg = sum/nsize;

#pragma omp parallel for num_threads(NUM_THRS)
    for(int n=0;n<nsize;n++)
		X_[n]-=avg;
}
//
void Poisson_Symmetric_7Point_CM::apply_A(const VECTOR& X, VECTOR& B )const{
	int             nsize = index3d_cm.     nsize;
	const int* index_nim1 = index3d_cm.index_nim1; const int* index_nip1 = index3d_cm.index_nip1;
	const int* index_njm1 = index3d_cm.index_njm1; const int* index_njp1 = index3d_cm.index_njp1;
	const int* index_nkm1 = index3d_cm.index_nkm1; const int* index_nkp1 = index3d_cm.index_nkp1;
    const double*     X_=X    .data;
	      double*     B_=B    .data;
	const double* Adiag_=Adiag.data;
	const double*  Aiph_=Aiph .data;
	const double*  Ajph_=Ajph .data;
	const double*  Akph_=Akph .data;
#pragma omp parallel for num_threads(NUM_THRS)
	for(int n=0;n<nsize;n++)
	{
		int nim1 = index_nim1[n], nip1 = index_nip1[n];
		int njm1 = index_njm1[n], njp1 = index_njp1[n];
		int nkm1 = index_nkm1[n], nkp1 = index_nkp1[n];
		double sum = Adiag_[n]*X_[n];
		if(nim1!=-1) sum+=Aiph_[nim1]*X_[nim1];
		if(njm1!=-1) sum+=Ajph_[njm1]*X_[njm1];
		if(nkm1!=-1) sum+=Akph_[nkm1]*X_[nkm1];
		if(nip1!=-1) sum+=Aiph_[n]*X_[nip1];
		if(njp1!=-1) sum+=Ajph_[n]*X_[njp1];
		if(nkp1!=-1) sum+=Akph_[n]*X_[nkp1];
		B_[n]=sum;
	}
}
//-------------------------------------------------------------------------------------------------
//  ILU : Dii + sum_{k=0:i-1}Lik/Dkk*                Uki = D0ii
// MILU : Dii + sum_{k=0:i-1}Lik/Dkk*(sum_{j=k+1:n-1}Ukj)= D0ii
//-------------------------------------------------------------------------------------------------
void Poisson_Symmetric_7Point_CM::calculate_MILU()
{
    Mdiag = Adiag;
	const double* Adiag_=Adiag.data;
	      double* Mdiag_=Mdiag.data;
	const double*  Aiph_=Aiph .data;
	const double*  Ajph_=Ajph .data;
	const double*  Akph_=Akph .data;
	int             nsize = index3d_cm.     nsize;
	const int* index_nim1 = index3d_cm.index_nim1; const int* index_nip1 = index3d_cm.index_nip1;
	const int* index_njm1 = index3d_cm.index_njm1; const int* index_njp1 = index3d_cm.index_njp1;
	const int* index_nkm1 = index3d_cm.index_nkm1; const int* index_nkp1 = index3d_cm.index_nkp1;
	for(int n=0;n<nsize;n++)
	{
		int nim1 = index_nim1[n], nip1 = index_nip1[n];
		int njm1 = index_njm1[n], njp1 = index_njp1[n];
		int nkm1 = index_nkm1[n], nkp1 = index_nkp1[n];
		double aiph = Aiph_[n];
		double ajph = Ajph_[n];
		double akph = Akph_[n];
		// ILU-MILU combination
		if(nip1!=-1) Mdiag_[nip1] -= aiph*aiph/Mdiag_[n];//*(aiph+.96*(ajph+akph));
		if(njp1!=-1) Mdiag_[njp1] -= ajph*ajph/Mdiag_[n];//*(ajph+.96*(aiph+akph));
		if(nkp1!=-1) Mdiag_[nkp1] -= akph*akph/Mdiag_[n];//*(akph+.96*(aiph+ajph));
	}
}

//
void Poisson_Symmetric_7Point_CM::PCG(const VECTOR& B, VECTOR& X, int NUM_THRS_, 
                            bool impose_sum_zero, 
							int max_it, double tolerance)
{
	NUM_THRS = NUM_THRS_;
	calculate_MILU();
	//
	int             nsize = index3d_cm.     nsize;
	const int* index_nim1 = index3d_cm.index_nim1; const int* index_nip1 = index3d_cm.index_nip1;
	const int* index_njm1 = index3d_cm.index_njm1; const int* index_njp1 = index3d_cm.index_njp1;
	const int* index_nkm1 = index3d_cm.index_nkm1; const int* index_nkp1 = index3d_cm.index_nkp1;
    const double* Adiag_=Adiag.data;
	const double*  Aiph_=Aiph .data;
	const double*  Ajph_=Ajph .data;
	const double*  Akph_=Akph .data;
	assert(B.size==nsize &&
           X.size==nsize);
	// pointers for faster acess
	VECTOR R (nsize); double*  R_=R.data;
    VECTOR P (nsize); double*  P_=P.data;
    VECTOR Z (nsize); double*  Z_=Z.data;  
    VECTOR& AP  = Z ; double* AP_=Z.data;
	      			  double*  X_=X.data; 
				const double*  B_=B.data;
    // set R=B-AX 
	apply_A(X,AP); 
	#pragma omp parallel for num_threads(NUM_THRS)
	for(int n=0;n<nsize;n++){ R_[n] = B_[n]-AP_[n];
							  Z_[n] = R_[n]       ;}

	if(impose_sum_zero){ projection_sumzero(R); projection_sumzero(Z); }
	// set Z=(1/M)R, P=Z
	apply_Minv(Z);

	#pragma omp parallel for num_threads(NUM_THRS)
	for(int n=0;n<nsize;n++) P_[n] = Z_[n];

	double rz = inner_product(Z,R); double rz0 = ABS(rz);
    // iteration
    for(int it=0;it<max_it;it++){
		if(ABS(rz)<=1E-14*rz0){ printf("MILU-PCG converged in %d iterations.\n",it); return; }
		apply_A(P,AP); if(impose_sum_zero) projection_sumzero(AP);
		double pAp = inner_product(P,AP);
		double a = rz/pAp;  
		#pragma omp parallel for num_threads(NUM_THRS) 
		for(int n=0;n<nsize;n++){
			X_[n]+=a* P_[n];
			R_[n]-=a*AP_[n];
			Z_[n] =   R_[n];}
		if(impose_sum_zero) projection_sumzero(Z); 
		apply_Minv (Z ); 
		double rz_new=inner_product(R,Z);
		double b = rz_new/rz; rz=rz_new; 
		#pragma omp parallel for num_threads(NUM_THRS) 
		for(int n=0;n<nsize;n++)
			P_[n] = Z_[n] + b*P_[n];
	}
    printf("MILU-PCG in %dx%dx%d FAILED in %d iterations\n",index3d_cm.isize,
		                                                    index3d_cm.jsize,
															index3d_cm.ksize,max_it);
	assert(false);
}
//
void Poisson_Symmetric_7Point_CM::apply_Minv(VECTOR& X)const
{
	int             isize = index3d_cm.     isize;
	int             jsize = index3d_cm.     jsize;
	int             ksize = index3d_cm.     ksize;
	int             nsize = index3d_cm.     nsize;
	const int* index_nim1 = index3d_cm.index_nim1; const int* index_nip1 = index3d_cm.index_nip1;
	const int* index_njm1 = index3d_cm.index_njm1; const int* index_njp1 = index3d_cm.index_njp1;
	const int* index_nkm1 = index3d_cm.index_nkm1; const int* index_nkp1 = index3d_cm.index_nkp1;
          double*     X_=X    .data;
	const double* Adiag_=Adiag.data;
	const double* Mdiag_=Mdiag.data;
	const double*  Aiph_=Aiph .data;
	const double*  Ajph_=Ajph .data;
	const double*  Akph_=Akph .data;
	const int* sum_nbeg = index3d_cm.sum_nbeg ;
	const int* sum_nsize= index3d_cm.sum_nsize;
	//---------------------------------------------------------------------
	// (L/D+I)X=X
	//---------------------------------------------------------------------
	for(int s=0;s<isize+jsize+ksize-2;s++)
	{
		int _nbeg = sum_nbeg [s];
		int _nsize= sum_nsize[s];
		#pragma omp parallel for num_threads(NUM_THRS)
		for(int _n=0;_n<_nsize;_n++)
		{
			int n = _nbeg+_n;
			int nim1 = index_nim1[n];
			int njm1 = index_njm1[n];
			int nkm1 = index_nkm1[n];
		
			double sum = 0;
			if(nim1!=-1) sum+=X_[nim1]/Mdiag_[nim1]*Aiph_[nim1];
			if(njm1!=-1) sum+=X_[njm1]/Mdiag_[njm1]*Ajph_[njm1];
			if(nkm1!=-1) sum+=X_[nkm1]/Mdiag_[nkm1]*Akph_[nkm1];
			X_[n]-=sum;
		}
	}

	//---------------------------------------------------------------------
	// (U+D)X=X
	//---------------------------------------------------------------------
	for(int s=isize+jsize+ksize-3;s>=0;s--)
	{
		int _nbeg = sum_nbeg [s];
		int _nsize= sum_nsize[s];
		#pragma omp parallel for num_threads(NUM_THRS)
		for(int _n=0;_n<_nsize;_n++)
		{
			int n = _nbeg+_n;
			int nip1 = index_nip1[n];
			int njp1 = index_njp1[n];
			int nkp1 = index_nkp1[n];
			double sum = 0;
			if(nip1!=-1) sum+=X_[nip1]*Aiph_[n];
			if(njp1!=-1) sum+=X_[njp1]*Ajph_[n];
			if(nkp1!=-1) sum+=X_[nkp1]*Akph_[n];
			X_[n]-=sum;
			X_[n]/=Mdiag_[n];
		}
	}
}
