#include <stdlib.h>
#include <math.h>
//#include <sys/time.h>
//#include <unistd.h>
#include "Poisson_Symmetric_7Point_CM.h"

const int N=150;
double get_phi(double x, double y, double z){ return SQR(sqrt(x*x+y*y)-2)+SQR(z)-1;}
double get_u  (double x, double y, double z){ return sin(x)*cos(y)*exp(-x*x-y*y);}
double get_w  (double x, double y, double z){ return 1+.5*sin(x+y+z); }
double get_rhs(double x, double y, double z){ return -exp(-y*y-x*x)*((4*sin(x)*y*sin(y)
                     +(4*sin(x)*y*y+(4*x*x-6)*sin(x)-4*x*cos(x))*cos(y))*sin(z+y+x)
                     +((-2*sin(x)*y-2*x*sin(x)+cos(x))*cos(y)-sin(x)*sin(y))*cos(z+y+x)+8*sin(x)*y*sin(y)
					 +(8*sin(x)*y*y+(8*x*x-12)*sin(x)-8*x*cos(x))*cos(y))/2;}
//------------------------------------------------------------------------------
// main
//------------------------------------------------------------------------------
int main( int argc, char *argv[])
{
	int NUM_THRS;
	if(argc<2) NUM_THRS=1;
	else       NUM_THRS=atoi(argv[1]);
	//------------------------------------------------------------------------------
	// test problem : -div(w*grad(U))=RHS in {Phi<0}
	//------------------------------------------------------------------------------
	GRID3D grid; grid.set_grid(-3.5,3.5,N,
							   -3.5,3.5,N,
							   -3.5,3.5,N); double h = grid.dx;
	VECTOR U  (N*N*N); U=0; const double eps=1E-5;
	VECTOR RHS(N*N*N);
	Poisson_Symmetric_7Point_CM A; A.set_size(N,N,N);
	int                       nsize = A.index3d_cm.nsize;
	const int*             n_to_ijk = A.index3d_cm.n_to_ijk;
	for(int n=0;n<nsize;n++)
	{
		int i = *n_to_ijk; n_to_ijk++;	double x=grid.x_fr_i(i);
		int j = *n_to_ijk; n_to_ijk++;	double y=grid.y_fr_j(j);
		int k = *n_to_ijk; n_to_ijk++;	double z=grid.z_fr_k(k);
		double phi    =get_phi(x  ,y  ,z  );
		double phi_im1=get_phi(x-h,y  ,z  ), phi_im2=get_phi(x-2*h,y    ,z    );
		double phi_ip1=get_phi(x+h,y  ,z  ), phi_ip2=get_phi(x+2*h,y    ,z    );
		double phi_jm1=get_phi(x  ,y-h  ,z), phi_jm2=get_phi(x    ,y-2*h,z    );
		double phi_jp1=get_phi(x  ,y+h  ,z), phi_jp2=get_phi(x    ,y+2*h,z    );
		double phi_km1=get_phi(x  ,y  ,z-h), phi_km2=get_phi(x    ,y    ,z-2*h);
		double phi_kp1=get_phi(x  ,y  ,z+h), phi_kp2=get_phi(x    ,y    ,z+2*h);
		double phixx=phi_ip1-2*phi+phi_im1;
		double phiyy=phi_jp1-2*phi+phi_jm1;
		double phizz=phi_kp1-2*phi+phi_km1;
		double phixx_ip1=phi_ip2-2*phi_ip1+phi, phixx_iph=MINMOD(phixx_ip1,phixx);
		double phixx_im1=phi_im2-2*phi_im1+phi, phixx_imh=MINMOD(phixx_im1,phixx);
		double phiyy_jp1=phi_jp2-2*phi_jp1+phi, phiyy_jph=MINMOD(phiyy_jp1,phiyy);
		double phiyy_jm1=phi_jm2-2*phi_jm1+phi, phiyy_jmh=MINMOD(phiyy_jm1,phiyy);
		double phizz_kp1=phi_kp2-2*phi_kp1+phi, phizz_kph=MINMOD(phizz_kp1,phizz);
		double phizz_km1=phi_km2-2*phi_km1+phi, phizz_kmh=MINMOD(phizz_km1,phizz);
		//------------------------------------------------------------------------------
		// outside the domain and near the boundary
		//------------------------------------------------------------------------------
		if(phi>-eps)
		{
			A.Adiag.data[n]= 1; A.Aiph.data[n]=0; 
			RHS    .data[n]= 0; A.Ajph.data[n]=0;
						        A.Akph.data[n]=0; 
		}
		//------------------------------------------------------------------------------
		// inside the domain 
		//------------------------------------------------------------------------------
		else
		{
			//------------------------------------------------------------------------------
			// distances from the center (x,y,z) to its neighboring nodes
			//------------------------------------------------------------------------------
			double dxm = h; if(phi_im1>0) dxm*=subcell_resolution(phi,phi_im1,phixx_imh); assert(dxm>1E-6);
			double dxp = h; if(phi_ip1>0) dxp*=subcell_resolution(phi,phi_ip1,phixx_iph); assert(dxp>1E-6);
			double dym = h; if(phi_jm1>0) dym*=subcell_resolution(phi,phi_jm1,phiyy_jmh); assert(dym>1E-6);
			double dyp = h; if(phi_jp1>0) dyp*=subcell_resolution(phi,phi_jp1,phiyy_jph); assert(dyp>1E-6);
			double dzm = h; if(phi_km1>0) dzm*=subcell_resolution(phi,phi_km1,phizz_kmh); assert(dzm>1E-6);
			double dzp = h; if(phi_kp1>0) dzp*=subcell_resolution(phi,phi_kp1,phizz_kph); assert(dzp>1E-6);
			//------------------------------------------------------------------------------
			// LHS : Gibou's symmetric discretization
			//------------------------------------------------------------------------------
			double wimh = get_w(x-h/2,y,z), wiph = get_w(x+h/2,y,z);
			double wjmh = get_w(x,y-h/2,z), wjph = get_w(x,y+h/2,z);
			double wkmh = get_w(x,y,z-h/2), wkph = get_w(x,y,z+h/2);
			A.Adiag.data[n] = (wimh/dxm+wiph/dxp+wjmh/dym+wjph/dyp+wkmh/dzm+wkph/dzp)/h;
			A.Aiph .data[n] = (        -wiph/dxp                                    )/h;
			A.Ajph .data[n] = (                          -wjph/dyp                  )/h;
			A.Akph .data[n] = (                                            -wkph/dzp)/h;
			//------------------------------------------------------------------------------
			// RHS
			//------------------------------------------------------------------------------
			RHS.data[n] = get_rhs(x,y,z);
			//------------------------------------------------------------------------------
			// Dirichlet boundary condition
			//------------------------------------------------------------------------------
			if(get_phi(x-h,y,z)>-eps)  RHS.data[n]+=wimh/dxm/h*get_u(x-dxm,y,z);
			if(get_phi(x+h,y,z)>-eps){ RHS.data[n]+=wiph/dxp/h*get_u(x+dxp,y,z); A.Aiph.data[n]=0; }
			if(get_phi(x,y-h,z)>-eps)  RHS.data[n]+=wjmh/dym/h*get_u(x,y-dym,z);	      
			if(get_phi(x,y+h,z)>-eps){ RHS.data[n]+=wjph/dyp/h*get_u(x,y+dyp,z); A.Ajph.data[n]=0; }
			if(get_phi(x,y,z-h)>-eps)  RHS.data[n]+=wkmh/dzm/h*get_u(x,y,z-dzm);	  
			if(get_phi(x,y,z+h)>-eps){ RHS.data[n]+=wkph/dzp/h*get_u(x,y,z+dzp); A.Akph.data[n]=0; }
		}
	}
	//------------------------------------------------------------------------------
	// solve the linear system
	//------------------------------------------------------------------------------
	printf("%d^3 grid with %d threads\n",N,NUM_THRS);
	//struct timeval start,end;
	//gettimeofday(&start,NULL);
	A.PCG(RHS,U,NUM_THRS);
	//gettimeofday(&end,NULL);
	//long seconds  = end.tv_sec  - start.tv_sec;     
	//long useconds = end.tv_usec - start.tv_usec;      
	//long mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;      
	//printf("Elapsed time: %ld milliseconds\n", mtime); 	

	ARRAY3D<double> U_(N,N,N);
	ARRAY3D<double> Phi(N,N,N);
	for(int i=0;i<N;i++)
	for(int j=0;j<N;j++)
	for(int k=0;k<N;k++){ U_(i,j,k)=U.data[A.index3d_cm.ijk_to_n[i][j][k]];
						  Phi(i,j,k)=get_phi(grid.x_fr_i(i),
							                 grid.y_fr_j(j),
											 grid.z_fr_k(k));}
	ARRAY2D<double>   U_sectionz;  U_.get_zcen(  U_sectionz);
	ARRAY2D<double> Phi_sectionz; Phi.get_zcen(Phi_sectionz);
	  U_sectionz.save_mfile("  U_sectionz_Dirichlet");
	Phi_sectionz.save_mfile("Phi_sectionz_Dirichlet");
	//------------------------------------------------------------------------------
	// error
	//------------------------------------------------------------------------------
	Statistics st;
	for(int i=1;i<N-1;i++){ double x=grid.x_fr_i(i);
	for(int j=1;j<N-1;j++){ double y=grid.y_fr_j(j);
	for(int k=1;k<N-1;k++){ double z=grid.z_fr_k(k);
		if(get_phi(x,y,z)<0)
			st.put_data(ABS(get_u(x,y,z)-U_(i,j,k)));
	}}}
	st.print_data("error");

	return 0;
}
