#include <stdlib.h>
#include <math.h>
#include "Poisson_Symmetric_7Point_CM.h"
#include "Levelsetmethod3D.h"

const int N=150;
const double radius = .5;	
double get_r  (double x, double y, double z){ return sqrt(x*x+y*y+z*z); }
double get_phi(double x, double y, double z){ double r=get_r(x,y,z); return r-radius;}
double get_u  (double x, double y, double z){ double r=get_r(x,y,z); return CUBE(SQR(r)-.25);}
double get_rhs(double x, double y, double z){ double r=get_r(x,y,z); return -((336*SQR(r)-120)*SQR(r)+9)/8.;}
//------------------------------------------------------------------------------
// main
//------------------------------------------------------------------------------
int main( int argc, char *argv[])
{
	int NUM_THRS;
	if(argc<2) NUM_THRS=1;
	else       NUM_THRS=atoi(argv[1]);
	//------------------------------------------------------------------------------
	// Heaviside function
	//------------------------------------------------------------------------------
	GRID3D grid; grid.set_grid(-1,1,N,
							   -1,1,N,
							   -1,1,N); double h = grid.dx;
	ARRAY3D<double> Phi (N  ,N  ,N  );
	ARRAY3D<double> Hiph(N-1,N  ,N  ); Hiph=0;
	ARRAY3D<double> Hjph(N  ,N-1,N  ); Hjph=0;
	ARRAY3D<double> Hkph(N  ,N  ,N-1); Hkph=0;
	for(int i=0;i<N;i++){ double x=grid.x_fr_i(i);
	for(int j=0;j<N;j++){ double y=grid.y_fr_j(j);
	for(int k=0;k<N;k++){ double z=grid.z_fr_k(k); Phi(i,j,k)=get_phi(x,y,z);}}}
	Levelsetmethod3D ls; 
	ls.grid=grid;
	ls.Phi =Phi ; 
	ls.discretize_heaviside_area_fraction(Hiph,Hjph,Hkph);
	//------------------------------------------------------------------------------
	// test problem : -div(H*grad(U))=RHS in {Phi<0}
	//------------------------------------------------------------------------------
	VECTOR U  (N*N*N); U=0; const double eps=1E-12;
	VECTOR RHS(N*N*N); RHS=0;
	Poisson_Symmetric_7Point_CM A; A.set_size(N,N,N);
	int                       nsize = A.index3d_cm.nsize;
	const int*             n_to_ijk = A.index3d_cm.n_to_ijk;
	for(int n=0;n<nsize;n++)
	{
		int i = *n_to_ijk; n_to_ijk++;	double x=grid.x_fr_i(i);
		int j = *n_to_ijk; n_to_ijk++;	double y=grid.y_fr_j(j);
		int k = *n_to_ijk; n_to_ijk++;	double z=grid.z_fr_k(k);
		//------------------------------------------------------------------------------
		// LHS 
		//------------------------------------------------------------------------------
		double wimh = (i==0)?0:(1-Hiph(i-1,j,k)), wiph = (i==N-1)?0:(1-Hiph(i,j,k));
		double wjmh = (j==0)?0:(1-Hjph(i,j-1,k)), wjph = (j==N-1)?0:(1-Hjph(i,j,k));
		double wkmh = (k==0)?0:(1-Hkph(i,j,k-1)), wkph = (k==N-1)?0:(1-Hkph(i,j,k));
		if(wimh<eps) wimh=0; if(wiph<eps) wiph=0;
		if(wjmh<eps) wjmh=0; if(wjph<eps) wjph=0;
		if(wkmh<eps) wkmh=0; if(wkph<eps) wkph=0;
		A.Adiag.data[n]= (wimh+wiph+wjmh+wjph+wkmh+wkph);
		A.Aiph .data[n]= (    -wiph                    );
		A.Ajph .data[n]= (              -wjph          );
		A.Akph .data[n]= (                        -wkph);
		//------------------------------------------------------------------------------
		// RHS
		//------------------------------------------------------------------------------
		RHS.data[n] = get_rhs(x,y,z)*h*h;
		//------------------------------------------------------------------------------
		// outside the domain
		//------------------------------------------------------------------------------
		if( A.Adiag.data[n]==0 )
		{
			A.Adiag.data[n]=1; RHS.data[n]=0;
		}
	}
	//------------------------------------------------------------------------------
	// solve the linear system
	//------------------------------------------------------------------------------
	printf("%d^3 grid with %d threads\n",N,NUM_THRS);
	//struct timeval start,end;
	//gettimeofday(&start,NULL);
	A.PCG(RHS,U,NUM_THRS,true);
	//gettimeofday(&end,NULL);
	//long seconds  = end.tv_sec  - start.tv_sec;     
	//long useconds = end.tv_usec - start.tv_usec;      
	//long mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;      
	//printf("Elapsed time: %ld milliseconds\n", mtime); 	
	//U.save("U.df3");

	ARRAY3D<double> U_(N,N,N);
	for(int i=0;i<N;i++)
	for(int j=0;j<N;j++)
	for(int k=0;k<N;k++) U_(i,j,k)=U.data[A.index3d_cm.ijk_to_n[i][j][k]];

	ARRAY2D<double> Usection(N,N); U_.get_zcen(Usection); Usection.save_mfile("Usectionz_Neumann");
	ARRAY2D<double> Phisection(N,N); Phi.get_zcen(Phisection); Phisection.save_mfile("Phisectionz_Neumann");
	//------------------------------------------------------------------------------
	// calculate the average 
	//------------------------------------------------------------------------------
	double uexct_avg=0;
	double uappr_avg=0; int count=0;
	for(int i=1;i<N-1;i++){ double x=grid.x_fr_i(i);
	for(int j=1;j<N-1;j++){ double y=grid.y_fr_j(j);
	for(int k=1;k<N-1;k++){ double z=grid.z_fr_k(k);
		if(get_phi(x,y,z)<0)
		{
			uexct_avg += get_u(x,y,z);
			uappr_avg += U_(i,j,k);
			count ++;
		}
	}}}
	uexct_avg/=count;
	uappr_avg/=count;
	//------------------------------------------------------------------------------
	// error
	//------------------------------------------------------------------------------
	Statistics st;
	for(int i=1;i<N-1;i++){ double x=grid.x_fr_i(i);
	for(int j=1;j<N-1;j++){ double y=grid.y_fr_j(j);
	for(int k=1;k<N-1;k++){ double z=grid.z_fr_k(k);
		if(get_phi(x,y,z)<0)
			st.put_data(ABS(get_u(x,y,z)-U_(i,j,k)-(uexct_avg-uappr_avg)));
	}}}
	st.print_data("error");

	return 0;
}
